@@ -2,6 +2,7 @@ import argparse
 import contextlib
 import collections
 import enum
+import errno
 import grp
 import hashlib
 import logging
@@ -134,11 +135,11 @@ def check_extension_modules():
         raise ExtensionModuleError
     if chunker.API_VERSION != '1.1_01':
         raise ExtensionModuleError
-    if compress.API_VERSION != '1.1_03':
+    if compress.API_VERSION != '1.1_06':
         raise ExtensionModuleError
     if borg.crypto.low_level.API_VERSION != '1.1_02':
         raise ExtensionModuleError
-    if platform.API_VERSION != platform.OS_API_VERSION != '1.1_02':
+    if platform.API_VERSION != platform.OS_API_VERSION or platform.API_VERSION != '1.1_03':
         raise ExtensionModuleError
     if item.API_VERSION != '1.1_02':
         raise ExtensionModuleError
@@ -464,25 +465,28 @@ def prune_split(archives, pattern, n, skip=[]):
     return keep
 
 
-def get_home_dir():
-    """Get user's home directory while preferring a possibly set HOME
-    environment variable
+def get_base_dir():
+    """Get home directory / base directory for borg:
+
+    - BORG_BASE_DIR, if set
+    - HOME, if set
+    - ~$USER, if USER is set
+    - ~
     """
+    base_dir = os.environ.get('BORG_BASE_DIR') or os.environ.get('HOME')
     # os.path.expanduser() behaves differently for '~' and '~someuser' as
     # parameters: when called with an explicit username, the possibly set
     # environment variable HOME is no longer respected. So we have to check if
     # it is set and only expand the user's home directory if HOME is unset.
-    if os.environ.get('HOME', ''):
-        return os.environ.get('HOME')
-    else:
-        return os.path.expanduser('~%s' % os.environ.get('USER', ''))
+    if not base_dir:
+        base_dir = os.path.expanduser('~%s' % os.environ.get('USER', ''))
+    return base_dir
 
 
 def get_keys_dir():
     """Determine where to repository keys and cache"""
 
-    xdg_config = os.environ.get('XDG_CONFIG_HOME', os.path.join(get_home_dir(), '.config'))
-    keys_dir = os.environ.get('BORG_KEYS_DIR', os.path.join(xdg_config, 'borg', 'keys'))
+    keys_dir = os.environ.get('BORG_KEYS_DIR', os.path.join(get_config_dir(), 'keys'))
     if not os.path.exists(keys_dir):
         os.makedirs(keys_dir)
         os.chmod(keys_dir, stat.S_IRWXU)
@@ -491,8 +495,7 @@ def get_keys_dir():
 
 def get_security_dir(repository_id=None):
     """Determine where to store local security information."""
-    xdg_config = os.environ.get('XDG_CONFIG_HOME', os.path.join(get_home_dir(), '.config'))
-    security_dir = os.environ.get('BORG_SECURITY_DIR', os.path.join(xdg_config, 'borg', 'security'))
+    security_dir = os.environ.get('BORG_SECURITY_DIR', os.path.join(get_config_dir(), 'security'))
     if repository_id:
         security_dir = os.path.join(security_dir, repository_id)
     if not os.path.exists(security_dir):
@@ -503,7 +506,7 @@ def get_security_dir(repository_id=None):
 
 def get_cache_dir():
     """Determine where to repository keys and cache"""
-    xdg_cache = os.environ.get('XDG_CACHE_HOME', os.path.join(get_home_dir(), '.cache'))
+    xdg_cache = os.environ.get('XDG_CACHE_HOME', os.path.join(get_base_dir(), '.cache'))
     cache_dir = os.environ.get('BORG_CACHE_DIR', os.path.join(xdg_cache, 'borg'))
     if not os.path.exists(cache_dir):
         os.makedirs(cache_dir)
@@ -513,11 +516,21 @@ def get_cache_dir():
             fd.write(textwrap.dedent("""
                 # This file is a cache directory tag created by Borg.
                 # For information about cache directory tags, see:
-                #       http://www.brynosaurus.com/cachedir/
+                #       http://www.bford.info/cachedir/spec.html
                 """).encode('ascii'))
     return cache_dir
 
 
+def get_config_dir():
+    """Determine where to store whole config"""
+    xdg_config = os.environ.get('XDG_CONFIG_HOME', os.path.join(get_base_dir(), '.config'))
+    config_dir = os.environ.get('BORG_CONFIG_DIR', os.path.join(xdg_config, 'borg'))
+    if not os.path.exists(config_dir):
+        os.makedirs(config_dir)
+        os.chmod(config_dir, stat.S_IRWXU)
+    return config_dir
+
+
 def to_localtime(ts):
     """Convert datetime object from UTC to local time zone"""
     return datetime(*time.localtime((ts - datetime(1970, 1, 1, tzinfo=timezone.utc)).total_seconds())[:6])
@@ -581,7 +594,7 @@ def dir_is_cachedir(path):
     """Determines whether the specified path is a cache directory (and
     therefore should potentially be excluded from the backup) according to
     the CACHEDIR.TAG protocol
-    (http://www.brynosaurus.com/cachedir/spec.html).
+    (http://www.bford.info/cachedir/spec.html).
     """
 
     tag_path = os.path.join(path, CACHE_TAG_NAME)
@@ -651,13 +664,15 @@ def format_line(format, data):
 
 def replace_placeholders(text):
     """Replace placeholders in text with their values."""
-    current_time = datetime.now()
+    from .platform import fqdn, hostname
+    current_time = datetime.now(timezone.utc)
     data = {
         'pid': os.getpid(),
-        'fqdn': socket.getfqdn(),
-        'hostname': socket.gethostname(),
-        'now': DatetimeWrapper(current_time.now()),
-        'utcnow': DatetimeWrapper(current_time.utcnow()),
+        'fqdn': fqdn,
+        'reverse-fqdn': '.'.join(reversed(fqdn.split('.'))),
+        'hostname': hostname,
+        'now': DatetimeWrapper(current_time.astimezone(None)),
+        'utcnow': DatetimeWrapper(current_time),
         'user': uid2user(os.getuid(), os.getuid()),
         'uuid4': str(uuid.uuid4()),
         'borgversion': borg_version,
@@ -732,11 +747,11 @@ def safe_timestamp(item_timestamp_ns):
     return datetime.fromtimestamp(t_ns / 1e9)
 
 
-def format_time(ts: datetime):
+def format_time(ts: datetime, format_spec=''):
     """
     Convert *ts* to a human-friendly format with textual weekday.
     """
-    return ts.strftime('%a, %Y-%m-%d %H:%M:%S')
+    return ts.strftime('%a, %Y-%m-%d %H:%M:%S' if format_spec == '' else format_spec)
 
 
 def isoformat_time(ts: datetime):
@@ -771,7 +786,7 @@ class OutputTimestamp:
         self.ts = ts
 
     def __format__(self, format_spec):
-        return format_time(self.ts)
+        return format_time(self.ts, format_spec=format_spec)
 
     def __str__(self):
         return '{}'.format(self)
@@ -1030,11 +1045,11 @@ class Location:
         """ + optional_archive_re, re.VERBOSE)              # archive name (optional, may be empty)
 
     def __init__(self, text=''):
-        self.orig = text
-        if not self.parse(self.orig):
-            raise ValueError('Location: parse failed: %s' % self.orig)
+        if not self.parse(text):
+            raise ValueError('Invalid location format: "%s"' % self.orig)
 
     def parse(self, text):
+        self.orig = text
         text = replace_placeholders(text)
         valid = self._parse(text)
         if valid:
@@ -1046,10 +1061,9 @@ class Location:
         if repo is None:
             return False
         valid = self._parse(repo)
-        if not valid:
-            return False
         self.archive = m.group('archive')
-        return True
+        self.orig = repo if not self.archive else '%s::%s' % (repo, self.archive)
+        return valid
 
     def _parse(self, text):
         def normpath_special(p):
@@ -1095,9 +1109,9 @@ class Location:
         return ', '.join(items)
 
     def to_key_filename(self):
-        name = re.sub('[^\w]', '_', self.path).strip('_')
+        name = re.sub(r'[^\w]', '_', self.path).strip('_')
         if self.proto != 'file':
-            name = re.sub('[^\w]', '_', self.host) + '__' + name
+            name = re.sub(r'[^\w]', '_', self.host) + '__' + name
         if len(name) > 100:
             # Limit file names to some reasonable length. Most file systems
             # limit them to 255 [unit of choice]; due to variations in unicode
@@ -1130,16 +1144,21 @@ class Location:
                                            path)
 
 
-def location_validator(archive=None):
+def location_validator(archive=None, proto=None):
     def validator(text):
         try:
             loc = Location(text)
-        except ValueError:
-            raise argparse.ArgumentTypeError('Invalid location format: "%s"' % text) from None
+        except ValueError as err:
+            raise argparse.ArgumentTypeError(str(err)) from None
         if archive is True and not loc.archive:
             raise argparse.ArgumentTypeError('"%s": No archive specified' % text)
         elif archive is False and loc.archive:
-            raise argparse.ArgumentTypeError('"%s" No archive can be specified' % text)
+            raise argparse.ArgumentTypeError('"%s": No archive can be specified' % text)
+        if proto is not None and loc.proto != proto:
+            if proto == 'file':
+                raise argparse.ArgumentTypeError('"%s": Repository must be local' % text)
+            else:
+                raise argparse.ArgumentTypeError('"%s": Repository must be remote' % text)
         return loc
     return validator
 
@@ -1582,11 +1601,32 @@ class ProgressIndicatorEndless:
 
 
 def sysinfo():
-    info = []
-    info.append('Platform: %s' % (' '.join(platform.uname()), ))
+    show_sysinfo = os.environ.get('BORG_SHOW_SYSINFO', 'yes').lower()
+    if show_sysinfo == 'no':
+        return ''
+
+    python_implementation = platform.python_implementation()
+    python_version = platform.python_version()
+    # platform.uname() does a shell call internally to get processor info,
+    # creating #3732 issue, so rather use os.uname().
+    try:
+        uname = os.uname()
+    except AttributeError:
+        uname = None
     if sys.platform.startswith('linux'):
-        info.append('Linux: %s %s %s' % platform.linux_distribution())
-    info.append('Borg: %s  Python: %s %s' % (borg_version, platform.python_implementation(), platform.python_version()))
+        try:
+            linux_distribution = platform.linux_distribution()
+        except:
+            # platform.linux_distribution() is deprecated since py 3.5 and removed in 3.7.
+            linux_distribution = ('Unknown Linux', '', '')
+    else:
+        linux_distribution = None
+    info = []
+    if uname is not None:
+        info.append('Platform: %s' % (' '.join(uname), ))
+    if linux_distribution is not None:
+        info.append('Linux: %s %s %s' % linux_distribution)
+    info.append('Borg: %s  Python: %s %s' % (borg_version, python_implementation, python_version))
     info.append('PID: %d  CWD: %s' % (os.getpid(), os.getcwd()))
     info.append('sys.argv: %r' % sys.argv)
     info.append('SSH_ORIGINAL_COMMAND: %r' % os.environ.get('SSH_ORIGINAL_COMMAND'))
@@ -1638,20 +1678,23 @@ class BaseFormatter:
 
 class ArchiveFormatter(BaseFormatter):
     KEY_DESCRIPTIONS = {
-        'name': 'archive name interpreted as text (might be missing non-text characters, see barchive)',
         'archive': 'archive name interpreted as text (might be missing non-text characters, see barchive)',
+        'name': 'alias of "archive"',
         'barchive': 'verbatim archive name, can contain any character except NUL',
         'comment': 'archive comment interpreted as text (might be missing non-text characters, see bcomment)',
         'bcomment': 'verbatim archive comment, can contain any character except NUL',
-        'time': 'time (start) of creation of the archive',
         # *start* is the key used by borg-info for this timestamp, this makes the formats more compatible
         'start': 'time (start) of creation of the archive',
+        'time': 'alias of "start"',
         'end': 'time (end) of creation of the archive',
         'id': 'internal ID of the archive',
+        'hostname': 'hostname of host on which this archive was created',
+        'username': 'username of user who created this archive',
     }
     KEY_GROUPS = (
-        ('name', 'archive', 'barchive', 'comment', 'bcomment', 'id'),
-        ('time', 'start', 'end'),
+        ('archive', 'name', 'barchive', 'comment', 'bcomment', 'id'),
+        ('start', 'time', 'end'),
+        ('hostname', 'username'),
     )
 
     @classmethod
@@ -1694,8 +1737,10 @@ class ArchiveFormatter(BaseFormatter):
         self.format = partial_format(format, static_keys)
         self.format_keys = {f[1] for f in Formatter().parse(format)}
         self.call_keys = {
-            'comment': partial(self.get_comment, rs=True),
-            'bcomment': partial(self.get_comment, rs=False),
+            'hostname': partial(self.get_meta, 'hostname', rs=True),
+            'username': partial(self.get_meta, 'username', rs=True),
+            'comment': partial(self.get_meta, 'comment', rs=True),
+            'bcomment': partial(self.get_meta, 'comment', rs=False),
             'end': self.get_ts_end,
         }
         self.used_call_keys = set(self.call_keys) & self.format_keys
@@ -1733,8 +1778,9 @@ class ArchiveFormatter(BaseFormatter):
             self._archive = Archive(self.repository, self.key, self.manifest, self.name)
         return self._archive
 
-    def get_comment(self, rs):
-        return remove_surrogates(self.archive.comment) if rs else self.archive.comment
+    def get_meta(self, key, rs):
+        value = self.archive.metadata.get(key, '')
+        return remove_surrogates(value) if rs else value
 
     def get_ts_end(self):
         return self.format_time(self.archive.ts_end)
@@ -2066,7 +2112,7 @@ class GenericDirEntry:
 
 def scandir_generic(path='.'):
     """Like os.listdir(), but yield DirEntry objects instead of returning a list of names."""
-    for name in sorted(os.listdir(path)):
+    for name in os.listdir(path):
         yield GenericDirEntry(path, name)
 
 
@@ -2081,8 +2127,21 @@ except ImportError:
         scandir = scandir_generic
 
 
+def scandir_keyfunc(dirent):
+    try:
+        return (0, dirent.inode())
+    except OSError as e:
+        # maybe a permission denied error while doing a stat() on the dirent
+        logger.debug('scandir_inorder: Unable to stat %s: %s', dirent.path, e)
+        # order this dirent after all the others lexically by file name
+        # we may not break the whole scandir just because of an exception in one dirent
+        # ignore the exception for now, since another stat will be done later anyways
+        # (or the entry will be skipped by an exclude pattern)
+        return (1, dirent.name)
+
+
 def scandir_inorder(path='.'):
-    return sorted(scandir(path), key=lambda dirent: dirent.inode())
+    return sorted(scandir(path), key=scandir_keyfunc)
 
 
 def clean_lines(lines, lstrip=None, rstrip=None, remove_empty=True, remove_comments=True):
@@ -2283,8 +2342,13 @@ def truncate_and_unlink(path):
     recover. Refer to the "File system interaction" section
     in repository.py for further explanations.
     """
-    with open(path, 'r+b') as fd:
-        fd.truncate()
+    try:
+        with open(path, 'r+b') as fd:
+            fd.truncate()
+    except OSError as err:
+        if err.errno != errno.ENOTSUP:
+            raise
+        # don't crash if the above ops are not supported.
     os.unlink(path)
 
 
